// you can use ANYTHING in java.util.*
// you can add as many helper methods as you want
// do not add any other instance or class variables
// feel free to change main... we will test your methods on the same and other input

import ontologyview.utils.FMAGraph;
import java.util.*;

public class FMAQueries {

	private static final String IS_CONTINUOUS_WITH = "continuous with";
	// this its own inverse
	private static final String HAS_SUBCLASS = ":DIRECT-SUBCLASSES";
	private static final String HAS_SUPERCLASS = ":DIRECT-SUPERCLASSES";
	// these two are inverses of each other
	private static final String IS_SUPERCLASS_OF = ":DIRECT-SUBCLASSES";
	private static final String IS_SUBCLASS_OF = ":DIRECT-SUPERCLASSES";
	// these two are inverses of each other
	private static final String HAS_PART = "part";
	private static final String IS_PART_OF = "part of";
	// these two are inverses of each other

	private final FMAGraph fma;

    public FMAQueries(String fname) {
		fma = new FMAGraph(fname);
	}

	// returns the string that is the inverse predicate for the given predicate.  probably not necessary.
	// throws an IllegalArgumentException if predicate or its inverse does not exist
	private String getInversePredicate(String predicate) {
		return fma.getInversePredicate(predicate);
	}

	// returns a list of objects, such that each object o in the list,
	// the (s,p,o) triple is in the knowledge base
	// throws an IllegalArgumentException if the subject does not exist
	// throws an IllegalArgumentException if the predicate does not exist
	// returns an empty list if there are no objects
	private List<String> getObjects(String subject, String predicate) {
		return fma.getObjects(subject,predicate);
	}

	public Collection<String> directSubclasses(String subject) {
		List l=getObjects(subject,HAS_SUBCLASS);
		Collections.sort(l);
		return l;
	}

	public Collection<String> transitiveParts(String subject) {
		throw new UnsupportedOperationException();
	}

	public Collection<String> certainOrgansInPart(String subject) {
		throw new UnsupportedOperationException();
	}

	public Collection<String> commonPartsOfKids(String superclass) {
		throw new UnsupportedOperationException();
	}

	public Collection<String> shortestPath(String src, String dst) {
		throw new UnsupportedOperationException();
	}

	public Collection<Collection<String>> allShortestPaths(String src, String dst) {
		throw new UnsupportedOperationException();
	}

	public static void printCollection(Collection<String> c) {
		System.out.print("[");
		for(Iterator<String> i=c.iterator(); i.hasNext(); ) {
			System.out.print(i.next());
			if (i.hasNext()) System.out.print(", ");
		}
		System.out.println("]");
	}

	public static void printCollectionOfCollections(Collection<Collection<String>> cc) {
		for(Collection<String> c:cc)
		    printCollection(c);
	}

	public static String checkSum(Collection<String> c) {
		long l = 0;
		for(String s : c)
			for(int i = 0; i < s.length(); i++)
				l = 37*l + s.charAt(i);
		String s = Long.toHexString(l).toUpperCase();
		while(s.length()<16) s="0"+s;
		return s;
	}

    public static void main(String[] args) {
		System.out.println("loading..");
		// change the filename to something absolute if you're having trouble
    	FMAQueries fmaq = new FMAQueries("fmaSuperCompressed.zip");

		Collection<String> c;
    	c = fmaq.directSubclasses("Human body");
    	printCollection(c); // prints [Female human body, Male human body]
  		System.out.println(fmaq.checkSum(c)); // prints 9B898BCBDCF78E87

    	c = fmaq.transitiveParts("Retina");
    	printCollection(c);// prints transitive closure of parts of the retina
  		System.out.println(c.size()); // prints 41
  		System.out.println(fmaq.checkSum(c)); // prints CA961EECB5E70C8D

    	c = fmaq.certainOrgansInPart("Abdomen");
    	printCollection(c); // prints certain abdominal organs
  		System.out.println(c.size()); // prints 19
  		System.out.println(fmaq.checkSum(c)); // prints DDEBF49B03DB2363

    	c = fmaq.commonPartsOfKids("Human body");
    	printCollection(c); // prints stuff in M&F human bodies but not Human body
  		System.out.println(c.size()); // prints 38
  		System.out.println(fmaq.checkSum(c)); // prints 7BC87AAED8A420DF

    	c = fmaq.shortestPath("Myocardial zone 4","Myocardial zone 7");
    	printCollection(c); // prints out one of the lines from problem 5

    	Collection<Collection<String>> cc = fmaq.allShortestPaths("Myocardial zone 4","Myocardial zone 7");
    	printCollectionOfCollections(cc); // prints out:
    // [Myocardial zone 4, Myocardial zone 3, Myocardial zone 2, Myocardial zone 8, Myocardial zone 7]
    // [Myocardial zone 4, Myocardial zone 3, Myocardial zone 9, Myocardial zone 8, Myocardial zone 7]
    // [Myocardial zone 4, Myocardial zone 5, Myocardial zone 6, Myocardial zone 12, Myocardial zone 7]
    // [Myocardial zone 4, Myocardial zone 5, Myocardial zone 11, Myocardial zone 12, Myocardial zone 7]
    // [Myocardial zone 4, Myocardial zone 10, Myocardial zone 9, Myocardial zone 8, Myocardial zone 7]
    // [Myocardial zone 4, Myocardial zone 10, Myocardial zone 11, Myocardial zone 12, Myocardial zone 7]
    }
}