package ontologyview.utils;

import java.io.File;
import java.io.PrintStream;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.zip.*;
// any, boolean, class, float, instance, integer, string, symbol


public class SuperCompressedFMA implements Serializable {
    public static final int ANY=0;
    public static final int BOOLEAN=1;
    public static final int CLASS=2;
    public static final int FLOAT=3;
    public static final int INSTANCE=4;
    public static final int INTEGER=5;
    public static final int STRING=6;
    public static final int SYMBOL=7;

	private String[] frames;
	private int[] framesMapStr;
	private int[] framesMapId;
	private int[] framesById;
	private int[] storage;

	private int[] intStore;
	private double[] doubleStore;
	private String[] stringStore;

    public Map<String,String> invSlotMap;
    public Map<String,String> parentSlotMap;
    public Map<String,Integer> slotTypes;
    public Map<Integer,Integer> slotIdTypes;

    public Map<String,Integer> strToInt;
    public Map<Integer,String> intToStr;

    public SuperCompressedFMA() {
        strToInt=new HashMap();
        intToStr=new HashMap();
        slotTypes=new HashMap();
        invSlotMap=new HashMap();
        parentSlotMap=new HashMap();
    }

	public int find(String s) {
		int idx=Arrays.binarySearch(frames,s);
		if (idx<0)  return -1;
		return framesMapStr[idx];
	}

	public int find(int idx) {
		idx=Arrays.binarySearch(framesById,idx);
		if (idx<0) return -1;
		return framesMapId[idx];
	}

	public Object getFirstSlotValue(String entity, String slot) {
		int k=getFirstValue(find(entity),strToInt.get(slot));
		Object o;
		if (k==0) o=null;
		else if (slotTypes.get(slot)==BOOLEAN) {
			if (k==1) o=Boolean.FALSE;
			if (k==2) o=Boolean.TRUE;
			else throw new IllegalStateException();
		} else if (slotTypes.get(slot)==INTEGER || slotTypes.get(slot)==INSTANCE || slotTypes.get(slot)==CLASS) {
			o=new Integer(intStore[k-1]);
		} else if (slotTypes.get(slot)==FLOAT) {
			o=new Double(doubleStore[k-1]);
		} else if (slotTypes.get(slot)==STRING || slotTypes.get(slot)==SYMBOL) {
			if (k<0) o=frames[-k-1];
			else o=stringStore[k-1];
		}
		else throw new RuntimeException();
		return o;
	}

	public Object[] getSlotValues(String entity, String slot) {
		int[] arr=getValues(find(entity),strToInt.get(slot));
//		System.out.println(entity+" "+slot+" "+arr.length);
//		if (arr.length>0) System.out.println(arr[0]);
		Object[] o=new Object[arr.length];
		for(int i=0;i<arr.length;i++) {
			if (arr[i]==0) o[i]=null;
			else if (slotTypes.get(slot)==BOOLEAN) {
				if (arr[i]==1) o[i]=Boolean.FALSE;
				if (arr[i]==2) o[i]=Boolean.TRUE;
				else { o[i]=null; System.out.println(entity+"@"+slot+": "+arr[i]);
//				throw new IllegalStateException();
				}
			} else if (slotTypes.get(slot)==INTEGER || slotTypes.get(slot)==INSTANCE || slotTypes.get(slot)==CLASS) {
				o[i]=new Integer(intStore[arr[i]-1]);
			} else if (slotTypes.get(slot)==FLOAT) {
				o[i]=new Double(doubleStore[arr[i]-1]);
			} else if (slotTypes.get(slot)==STRING || slotTypes.get(slot)==SYMBOL) {
				if (arr[i]<0) o[i]=getName(-arr[i]-1);
				else o[i]=stringStore[arr[i]-1];
			}
			else throw new RuntimeException();
		}
		return o;
	}

	public int getSlotType(String slot) {
		return slotTypes.get(slot);
	}

	public Iterator<String> allSlots(final String e) {
		return new Iterator() {
			Iterator<Integer> it=(find(e)==-1)?null:allSlots(find(e));
		    public boolean hasNext() {
				return it==null?false:it.hasNext();
		    }
		    public String next() {
				return intToStr.get(it.next());
			}
			public void remove() { throw new UnsupportedOperationException(); }
		};
	}

	public Iterator<String> allDefinedNonRelSlots(final String e) {
		return new Iterator() {
			Iterator<Integer> it=(find(e)==-1)?null:allDefinedNonRelSlots(find(e));
		    public boolean hasNext() {
				return it==null?false:it.hasNext();
		    }
		    public String next() {
				return intToStr.get(it.next());
			}
			public void remove() { throw new UnsupportedOperationException(); }
		};
	}

	public Iterator<String> allNonEmptyRelations(final String e) {
		return new Iterator() {
			Iterator<Integer> it=(find(e)==-1)?null:allNonEmptyRelations(find(e));
		    public boolean hasNext() {
				return it==null?false:it.hasNext();
		    }
		    public String next() {
				return intToStr.get(it.next());
			}
			public void remove() { throw new UnsupportedOperationException(); }
		};
	}

	static int[] slIdx,con,conIdx;
	static Object[] sl;
	static int[] flags=new int[4];

    public static void main(String[] args) {
       CompressedFMA ret = null;
       FileInputStream fis = null;
       ObjectInputStream in = null;
       try {
//         fis = new FileInputStream(name);
		ZipFile zf=new ZipFile("fmaCompressed.zip");
//         ZipInputStream zi=new ZipInputStream(name);
         ZipEntry ze=(ZipEntry)zf.entries().nextElement();
         System.out.println(ze.getName());
         in = new ObjectInputStream(zf.getInputStream(ze));
//         in = new ObjectInputStream(fis);
         ret = (CompressedFMA)in.readObject();
         in.close();
       } catch(Exception ex) {
         ex.printStackTrace();
       }

		SuperCompressedFMA z = new SuperCompressedFMA();

        z.strToInt=ret.strToInt;
        z.intToStr=ret.intToStr;
        z.slotTypes=ret.slotTypes;
        z.invSlotMap=ret.invSlotMap;
        z.parentSlotMap=ret.parentSlotMap;
		z.slotIdTypes=new HashMap();

		for(Iterator<String> si=z.slotTypes.keySet().iterator(); si.hasNext(); ) {
			String s=si.next();
			z.slotIdTypes.put(z.strToInt.get(s),z.slotTypes.get(s));
		}

		z.frames=new String[ret.frames.size()];
		z.framesById=new int[ret.frames.size()];
		z.framesMapStr=new int[ret.frames.size()];
		z.framesMapId=new int[ret.frames.size()];
		HashMap<String,Integer> h=new HashMap();

		int k=0;
		int mem=0;
		for(Iterator<FMAItem> i=ret.frames.values().iterator(); i.hasNext(); ) {
			FMAItem f=i.next();
			z.frames[k]=f.getName();
			z.framesById[k]=f.getID();
			mem+=4; // flags
			mem+=2; // lengths
			mem+=1; // name
			mem+=sl.length+slIdx.length+con.length+conIdx.length;
			k++;
		}
		Arrays.sort(z.framesById);
		int lastInt=0,lastStr=0,lastObj=0,lastDouble=0;

		z.storage=new int[mem];
		for(Iterator<FMAItem> i=ret.frames.values().iterator(); i.hasNext(); ) {
			FMAItem f=i.next();
			String s=f.getName();
			for(int j=0;j<sl.length;j++) {
				if (sl[j] instanceof Boolean) {
				} else if (sl[j] instanceof Integer) {
					lastInt++;
				} else if (sl[j] instanceof Double) {
					lastDouble++;
				} else if (sl[j] instanceof String) {
					if (z.find((String)sl[j])==-1) {
						Integer qq=h.get((String)sl[j]);
						if (qq==null) {
//							z.stringStore[lastStr]=(String)sl[j];
							h.put((String)sl[j],lastStr);
//System.out.println(lastStr+": "+sl[j]);
							lastStr++;
						}
					}
				} else {
					if (sl[j]!=null)
						throw new RuntimeException();
				}
			}
		}
		z.intStore=new int[lastInt];
		z.stringStore=new String[lastStr];
		z.doubleStore=new double[lastDouble];
		lastInt=lastStr=lastDouble=0;
		k=0;
		mem=0;
		for(Iterator<FMAItem> i=ret.frames.values().iterator(); i.hasNext(); ) {
			FMAItem f=i.next();
			String s=f.getName();
			z.framesMapStr[k]=mem;
			int q=Arrays.binarySearch(z.framesById,f.getID());
			z.framesMapId[q]=mem;
			for(int j=0;j<4;j++)
				z.storage[mem++]=flags[j];
			z.storage[mem++]=slIdx.length+65536*sl.length;
			z.storage[mem++]=conIdx.length+65536*con.length;
			z.storage[mem++]=k;
			for(int j=0;j<slIdx.length;j++)
				z.storage[mem++]=slIdx[j];
			for(int j=0;j<sl.length;j++) {
				int idx=0;
				if (sl[j] instanceof Boolean) {
					if (sl[j].equals(Boolean.FALSE))
						idx=1;
					else idx=2;
				} else if (sl[j] instanceof Integer) {
					z.intStore[lastInt]=(Integer)sl[j];
					idx=1+lastInt++;
				} else if (sl[j] instanceof Double) {
					z.doubleStore[lastDouble]=(Double)sl[j];
					idx=1+lastDouble++;
				} else if (sl[j] instanceof String) {
					if (z.find((String)sl[j])==-1) {
//System.out.println(sl[j]);
						Integer qq=h.get((String)sl[j]);
						if (qq==lastStr) {
							z.stringStore[lastStr]=(String)sl[j];
							idx=1+lastStr++;
						} else {
							idx=qq+1;
						}
					} else
						idx=-z.find((String)sl[j])-1;
				} else {
					idx=0;
				}
				z.storage[mem++]=idx;
			}
			for(int j=0;j<conIdx.length;j++)
				z.storage[mem++]=conIdx[j];
			for(int j=0;j<con.length;j++)
				z.storage[mem++]=con[j];
			k++;
		}

                FileOutputStream fos = null;
                ObjectOutputStream out = null;
                System.out.println("storage: "+z.storage.length);
                System.out.println("intStore: "+z.intStore.length);
                System.out.println("doubleStore: "+z.doubleStore.length);
                int tot=0;
                for(int kk=0;kk<z.stringStore.length;kk++) tot+=z.stringStore[kk].length();
                System.out.println("stringStore: "+z.stringStore.length+" "+tot);
                tot=0;
                for(int kk=0;kk<z.frames.length;kk++) tot+=z.frames[kk].length();
                System.out.println("names: "+z.frames.length+" "+tot);
                System.out.println("maps: "+z.frames.length*4);

                try
                {
                   fos = new FileOutputStream("fmaSuperCompressed.bin");
                   out = new ObjectOutputStream(fos);
                   out.writeObject(z);
                   out.close();
                 }
                 catch(IOException ex)
                 {
                   ex.printStackTrace();
                 }
/*
                 HashMap<String,Integer> t=new HashMap();
                 for(int i=0;i<z.frames.length;i++) {
					String[] ss=z.frames[i].split("[\\s]");
					for(int j=0;j<ss.length;j++) {
						String s=ss[j].toLowerCase().trim();
						Integer q=t.get(s);
						if (q==null) t.put(s,0);
						else t.put(s,q+1);
					}
				 }
				 Combo[] t2=new Combo[t.size()];
				 int q=0;
				 for(String s:t.keySet()) {
				 	t2[q]=new Combo(-t.get(s),s);
				 	q++;
				}
				Arrays.sort(t2);
				 PrintStream p;
				 try {
				 	p=new PrintStream(new File("freq.txt"));
			 	 } catch (Exception e) {throw new RuntimeException(e);}
			 	 for(int j=0;j<t2.length;j++)
					 p.println(-t2[j].count+" "+t2[j].name);
*/
    }

	private static class Combo implements Comparable {
		int count;
		String name;
		Combo(int count, String name) {this.count=count; this.name=name;}
		public int compareTo(Object o) {
			Combo c=(Combo)o;
			return count-c.count;
		}
	}

    public static SuperCompressedFMA loadFromFile(String name) {
       SuperCompressedFMA ret = null;
       FileInputStream fis = null;
       ObjectInputStream in = null;
       try {
//         fis = new FileInputStream(name);
		ZipFile zf=new ZipFile(name);
//         ZipInputStream zi=new ZipInputStream(name);
         ZipEntry ze=(ZipEntry)zf.entries().nextElement();
         System.out.println(ze.getName());
         in = new ObjectInputStream(zf.getInputStream(ze));
//         in = new ObjectInputStream(fis);
         ret = (SuperCompressedFMA)in.readObject();
         in.close();
       } catch(Exception ex) {
         ex.printStackTrace();
       }
       return ret;
    }






    private static int[] EMPTY=new int[0];

    public String getName(int base) {
		return frames[storage[base+6]];
	}
/*
    public int getID(int base) {
		return getFirstValue(base,FMAID);
	}*/

    public boolean hasSlot(int base, int i) {
        int j=0;
        while(i>=32) {
            i-=32;
            j++;
        }
        return (storage[base+j]&(1<<i))!=0;
    }

	public Iterator<Integer> allSlots(final int base) {
		return new Iterator() {
			int word,val=storage[base],c;
		    int ret;
		    boolean ok;
		    public boolean hasNext() {
				while(!ok && word<4) {
//				System.out.println(word+" "+c+" "+val);
					ok=((val&1)!=0);
					val=val>>>1;
					if (ok) {
						ret=c+32*word;
					}
					c++;
					if (val==0) { word++; if (word>=4) break;
						val=storage[base+word]; c=0; }
				}
//				System.out.println(c+" "+word);
				return ok;
		    }
		    public Integer next() {
				if (!ok) hasNext();
				if (!ok) throw new NoSuchElementException();
				ok = false;
				return ret;
			}
		    public void remove() { throw new UnsupportedOperationException(); }
		};
	}

	public Iterator<Integer> allDefinedNonRelSlots(final int base) {
		return new Iterator() {
			Iterator<Integer> it=allSlots(base);
		    int ret;
		    boolean ok;
		    public boolean hasNext() {
				int n;
			    while(!ok && it.hasNext()) {
					ret=it.next();
					int[] foo=getValues(base,ret);
					ok=foo!=null && foo.length!=0;
				}
				return ok;
		    }
		    public Integer next() {
				if (!ok) hasNext();
				if (!ok) throw new NoSuchElementException();
				ok = false;
				return ret;
			}
		    public void remove() { throw new UnsupportedOperationException(); }
		};
	}

	public int conIdxL(int base) {
		return storage[base+5]%65536;
	}

	public int conL(int base) {
		return storage[base+5]/65536;
	}

	public int slIdxL(int base) {
		return storage[base+4]%65536;
	}

	public int slL(int base) {
		return storage[base+4]/65536;
	}

	public int conIdxB(int base) {
		return slB(base)+slL(base);
	}

	public int conB(int base) {
		return conIdxB(base)+conIdxL(base);
	}

	public int slIdxB(int base) {
		return base+7;
	}

	public int slB(int base) {
		return slIdxB(base)+slIdxL(base);
	}

	public Iterator<Integer> allNonEmptyRelations(final int base) {
		return new Iterator() {
			int idx;
		    int ret;
		    boolean ok;
		    public boolean hasNext() {
				int b=conIdxB(base);
				int l=conIdxL(base);
				int q=strToInt.get("regional part");
//				System.out.println("reg part: "+q);
			    while (!ok && idx<l) {
					int k=storage[b+idx++];
					ret=k&0xFFFF;
					int n=k>>>16;
					int len;
					if (idx==l) {
						len=conL(base)-n;
					} else {
						int k2=storage[b+idx];
						len=(k2>>>16)-n;
					}
					if (len!=0) ok=true;
//					System.out.println("idx: "+(idx-1)+"("+n+" "+len+" "+ret+")");
				}
//				System.out.println("???: "+ret+" "+intToStr.get(ret));
				return ok;
		    }
		    public Integer next() {
				if (!ok) hasNext();
				if (!ok) throw new NoSuchElementException();
				ok = false;
				return ret;
			}
		    public void remove() { throw new UnsupportedOperationException(); }
		};
	}

	public int[] getValues(int base, int src) {
//       System.out.println("searching: "+getName(base));
       int i=0;
        int b=slIdxB(base);
        int l=slIdxL(base);
        while(i<l) {
			int k=storage[b+i++];
            int slot=k&0xFFFF;
            int n=k>>>16;
            if (slot==src) {
				int len;
				if (i==l) len=slL(base)-n;
				else len=(storage[b+i]>>>16)-n;
                int[] ret=new int[len];
                int b2=slB(base);
                for(int j=0;j<len;j++) {
                    ret[j]=storage[b2+n+j];
                }
//                System.out.println("found! "+len);
                return ret;
            }
        }
		int[] foo=getConnectivity(base,0);
		int k=0;
		if (foo==null || foo.length==0) return EMPTY;
        if (foo.length>1) {
			int max=-1;
			int[] bar=EMPTY;
			while(k<foo.length) { // HACKHACK
				int b2=find(foo[k]);
				int[] tmp=null;
				if (b2!=-1) tmp=getValues(b2,src);
				if (tmp!=null && tmp.length>max) { max=tmp.length;
					bar=tmp; }
				k++;
			}
			return bar;
		} else {
			int b2=find(foo[0]);
			return b2==-1?EMPTY:getValues(b2,src);
		}
	}

	public int getFirstValue(int base, int src) {
        int i=0;
        int b=slIdxB(base);
        int l=slIdxL(base);
        while(i<l) {
			int k=storage[b+i++];
            int slot=k&0xFFFF;
            int n=k>>>16;
            if (slot==src) {
                return storage[slB(base)+n];
            }
        }
		int[] foo=getConnectivity(base,0);
		int k=0;
		if (foo==null || foo.length==0) return 0;
        if (foo.length>1) {
			int bar=0;
			while(k<foo.length) { // HACKHACK
				int b2=find(foo[k]);
				int tmp=b2==-1?null:getFirstValue(b2,src);
				if (tmp!=0) {
					if (bar!=0) {
						System.out.println("multiple inherited values!!!");
					}
					bar=tmp;
				}
				k++;
			}
			return bar;
		} else {
			int b2=find(foo[0]);
			return b2==-1?0:getFirstValue(b2,src);
		}
	}

    public int[] getConnectivity(int base, int src) {
        int i=0;
        int b=conIdxB(base);
        int l=conIdxL(base);
        while(i<l) {
			int k=storage[b+i++];
            int slot=k&0xFFFF;
            int n=k>>>16;
            if (slot==src) {
				int len;
				if (i==l) len=conL(base)-n;
				else len=(storage[b+i]>>>16)-n;
                int[] ret=new int[len];
                int b2=conB(base);
                for(int j=0;j<len;j++) {
                    ret[j]=storage[b2+n+j];
                }
                return ret;
            }
        }
        return null;
    }

    public int getConnectivityCount(int base, int src) {
        int i=0;
        int b=conIdxB(base);
        int l=conIdxL(base);
        while(i<l) {
			int k=storage[b+i++];
            int slot=k&0xFFFF;
            int n=k>>16;
            if (slot==src) {
				if (i==l) n=conL(base)-n;
				else n=(storage[b+i]>>16)-n;
                return n;
			}
        }
        return 0;
    }
}